#ifndef LocalVField_h
#define LocalVField_h

#include "capd/capdlib.h"
using namespace capd;

//////////////////////////////////
// class: LocalVField
//
// Purpose: The aim of the class is to consider a vector field in local coordinates in a neighbourhood of
// a fixed point. 
//
// The local coordinates "p" are: 
//    q = midPoint + C*p
// where "q" are the original coordinates of the ODE.
//
// Members:
//         F    - the vector field in original coordinates
//   midPoint   - a zero in local coordinates corresponds fo the midPoint in original coordinates
//         C    - Matrix which gives the local coordinates.
//      invC    - The bound on the inverse of C
//
// Derivation of the local vector field:
//    q = midPoint + C*p
//    p = invC*(q - midPoint)
//
//    q'= F(q)
//    p'= invC*q' = invC*F(q) = invC*F(midPoint + C*p)  
// The local vector field is:
//    G(p) = invC*F(midPoint + C*p) 
//    DG(p)= invC*DF(midPoint + C*p)*C
class LocalVField
{ 
public:
	IMap *F;
	IVector midPoint;
	IMatrix C;
	IMatrix invC;
	
	LocalVField(IMap &F, IVector midPoint,IMatrix C);
	IMatrix derivative(const IVector &p) const {return invC*(*F)[midPoint+C*p]*C;}
	IVector image(const IVector &p) const {return invC*(*F)(midPoint + C*p);}
	IMatrix operator[](const IVector &p) const {return derivative(p);}
	IVector operator()(const IVector &p) const {return image(p);}
};

#endif
