#include "WuEnclosure.h"
#include "ratecond.h"
#include "isolatingBlock.h"
#include "utils.h"
using namespace capd::matrixAlgorithms;

///////////////////////
// The fixed point is established using the interval Newton method
IVector FixedPoint(LocalVField &f)
{
	IVector x(3);
	for(int i=0;i<10;i++) x=midVector(x) - gauss(f[x],f(midVector(x)));
	x=midVector(x)+1.1*(x-midVector(x));
	int prec=14;
	if((x-midVector(x)).euclNorm()<pow(10.0,-prec)) for(int i=0;i<3;i++) x[i]=x[i].mid() + pow(10.0,-prec)*interval(-1.0,1.0);
	
	IVector N=midVector(x) - gauss(f[x],f(midVector(x)));

	if(! subsetInterior(N,x)) abort("Error in establishing the fixed point enclosure. Aborting program");
	return x;
}

///////////////////////
// This functions computes a bound on the derivative of the vector field on the set:
//   Bu x Bs
// and verifies if the rate conditions are satisfied on that set.
// If not, we throw an error and terminate the program.
// We subdivide the x-coordinate (the unstable coordinate) into many fragments to obtain
// sharper bounds for the derivative of the vector field.
// If the dimension of x is 1 we have 100 subdivisions.
// If the dimension of x is 2 we have 100*100 subdivisions.
void WuEnclosure::EstablishManifold(IVector Bu, IVector Bs)
{		
	fixedPoint=FixedPoint(*f);
	
	int n=domain.dimension();
	IVector q(3);
	for(int i=0;i<n;i++) q[i]=domain[i];
	for(int i=n;i<3;i++) q[i]=Bs[i-n];
	IMatrix DF=(*f)[fixedPoint];
	if(n==1)
	{
		IVector u=q;
		int N=100;
		for(int i=0;i<N;i++)
		{
			u[0]=part(q[0],N,i);
			DF=intervalHull(DF,(*f)[u]);
		} 
	}
	if(n==2)
	{
		IVector u=q;
		int N=100;
		for(int i=0;i<N;i++)
			for(int j=0;j<10;j++) 
			{
				u[0]=part(q[0],N,i);
				u[1]=part(q[1],N,j);
				DF=intervalHull(DF,(*f)[u]);
			}
	}
	if(satisfiesRateConditions(DF,L,n)==0) abort("unable to verify the enclosure of Wu. ABORTING PROGRAM. ");
	if(satisfiesIsolatingBlockCondition(*f,DF,n,domain[0].right(),Bs[0].right())==0) abort("unable to establish isolating block. ");
}


WuEnclosure::WuEnclosure(LocalVField &f_, IVector domain_, interval L_)
{
	f=&f_;
	domain=domain_;
	L=L_;
	int n=domain.dimension();
	IVector Bs(3-n);
	for(int i=n;i<3;i++) Bs[i-n]=L*(domain.euclNorm())*interval(-1.0,1.0);
	EstablishManifold(domain,Bs);
}

WuEnclosure::WuEnclosure(LocalVField &f_, IVector Bu, interval L_,IVector Bs)
{
	f=&f_;
	domain=Bu;
	L=L_;
	EstablishManifold(Bu,Bs);
}

C0Rect2Set WuEnclosure::image(const IVector &x) const
{
	IVector b=localImage(x);
	
	C0Rect2Set X((f->midPoint)+(f->C)*midVector(b),f->C,b-midVector(b));
	return X;
}

IVector WuEnclosure::localImage(const IVector &x) const
{	
	IVector b(3);
	int n=domain.dimension();
	IVector xfixedPoint(n); // this will be the the projection ont the x coordinate 
	for(int i=0;i<n;i++)
	{
		b[i]=x[i];
		xfixedPoint[i] = fixedPoint[i]; 
	} 
	for(int i=n;i<3;i++) b[i]=fixedPoint[i]+L*((x-xfixedPoint).euclNorm())*interval(-1.0,1.0);
	
	return b;
}
	