#include "capd/capdlib.h"
#include "utils.h"
#include "LocalVField.h"
#include "WuEnclosure.h"
#include "ratecond.h"
#include "eigenvector.h"
#include <iostream>

using namespace std;
using namespace capd;
using namespace capd::matrixAlgorithms;

////////////////////////////////
// Function: roughFixedPoint()
//
// Purpose: This function computes a fixed point candidate by solving
//    f(x)=0
// using the Newton method.
// Note: This is not a rigorous computation. It only serves the purpose of getting an initial guess,
// that can be validated using the interval Newton method.
IVector roughFixedPoint(IMap &f,IVector candidate)
{
	IVector x=candidate;
	for(int i=0;i<10;i++) x=midVector(x) - gauss(f[x],f(midVector(x)));
	return midVector(x);
}

////////////////////////////////
// Function: roughFixedPoint()
//
// Purpose: This function computes a fixed point by solving
//    f(x)=0
// using the interval Newton method.
IVector fixedPoint(IMap &f,IVector candidate)
{
	IVector x=roughFixedPoint(f,candidate);
	
	x=midVector(x)+1.1*(x-midVector(x)); // blowup of initial guess
	int prec=9;
	if((x-midVector(x)).euclNorm()<pow(10.0,-prec)) for(int i=0;i<3;i++) x[i]=x[i].mid() + pow(10.0,-prec)*interval(-1.0,1.0);
	
	IVector N=midVector(x) - gauss(f[x],f(midVector(x)));

	if(! subsetInterior(N,x)) abort("Error in establishing the fixed point enclosure. Aborting program");
	return x;
}

//////////////////////////////
// Function: computeEigenvalues()
//
// Purpose: This function computes and displays eigenvalues of the Jacobian of a
// vector field f, at a point x.
// The aim is simply to display these values in order to verify hypebolicity. They are not used
// anywhere else than that in the program.
//
// Method: We first compute an initial guess on the eigenvalue and eigenvector nonrigorously.
// This is then converted to a set enclosure and validated using:
//   containsEigenvalueAndEigenvector()
// from eigenvector.h/.cpp
void computeEigenvalues(IMap &f,IVector x)
{
	// ---------------------------
	// computation of the candidates for eigenvalues and eigenvectors:
	IMatrix A=f[x];
	DMatrix B(3,3);
	for(int i=0;i<3;i++)
		for(int k=0;k<3;k++) B[i][k]=A[i][k].mid().leftBound();
  	DVector rE(3), iE(3);         	// real and imaginary parts of eigenvalues
  	DMatrix rVec(3,3), iVec(3,3); 	// real and imaginary parts of eigenvectors
  	computeEigenvaluesAndEigenvectors(B,rE,iE,rVec,iVec);
	// end of computation of the candidates for eigenvalues and eigenvectors.
	// ---------------------------
	
    IVector xRe(3),xIm(3);
	double r=pow(10.0,-6.0); // this is the factor by which we "blow up" the initial guess
	
	for(int k=0;k<3;k++) 
	{
		for(int i=0;i<3;i++) xRe[i] = rVec[i][k]+r*interval(-1.0,1.0); // blowing up the initial guess
		for(int i=0;i<3;i++) xIm[i] = iVec[i][k]+r*interval(-1.0,1.0); // blowing up the initial guess
    
	    interval rho=rE[k]+r*interval(-1.0,1.0); // blowing up the initial guess
	    interval omega=iE[k]+r*interval(-1.0,1.0); // blowing up the initial guess
		
		
		if(abs(omega) < 2.0*r) // this means that it is likely that the eigenvalue is real, hence below we check for a real eigenvalue:
		{
			// If containsEigenvalueAndEigenvector(A,rho,xRe) returns 1, then this ensures that
			// the eigenvector and eigenvalue is real:
			if(containsEigenvalueAndEigenvector(A,rho,xRe)==0) abort("Problem with computing eigenvalues. Aborting program.");
			cout << "Eigenvalue " << k+1 << " Re = " << rho << "    Im = 0.0 " << endl;
		}
		else // we check for a complex eigenvalue:
		{
			if(containsEigenvalueAndEigenvector(A,rho,omega,xRe,xIm)==0) abort("Problem with computing eigenvalues. Aborting program.");
			cout << "Eigenvalue " << k+1 << " Re = " << rho << "    Im = " << omega << endl;
		}
	}
	
}

/////////////////////////////
// Function: writeOutLinearCoordChange()
//
// Purpose: This function writes out a coordinate change to linear Jordan form of the 
// Jacobian of a vector field at a point.
// The purpose is just to display the coordinate change. This function is not part of the proof,
// and is not used by any other routines.
void writeOutLinearCoordChange()
{
	string vFieldWu1="par:a,b,F,G;var:X,Y,Z;fun:Y^2+Z^2+a*X-a*F,-X*Y+b*X*Z+Y-G,-b*X*Y-X*Z+Z;";
	DMap f(vFieldWu1);  f.setParameter("a",0.25);  f.setParameter("b",4.0);  f.setParameter("F",4.0);
	f.setParameter("G",0.0752761125);
	DVector x(3);
	x[0]=4.0;
	for(int i=0;i<10;i++) x=x - gauss(f[x],f(x));
	DMatrix C=f[x];
	
  	DVector rE(3), iE(3);         	// real and imaginary parts of eigenvalues
  	DMatrix rVec(3,3), iVec(3,3); 	// real and imaginary parts of eigenvectors

  	computeEigenvaluesAndEigenvectors(C,rE,iE,rVec,iVec);
	cout << rVec << endl;
	cout << iVec << endl;
}

/////////////////////////////
// Function: paramSet()
//
// Purpose: This function returns a point of the form:
//    Y = X0 + C*R
// That encloses a manifold 'wu'. The objective is that the point Y
// is to be in the space extented by the parameter. Wet we do is 
// to construct an enclosure so that Y contains
//   midPoint + C*(G,wu(x))
// where wu(x) is the enclosure of the unstable manifold in local coordinates.
//
// Important: The matrix C should be chosen so that on the last three coordinates
// it agrees with the local coordinates of wu. On the first coordinate C should be the identity.
C1Rect2Set paramSet(WuEnclosure &wu,IMatrix &C,IVector x,interval G)
{
	x=wu.localImage(x);
	IVector X(4);
	for(int i=0;i<3;i++) X[i+1]=x[i];
	IVector X0(4);
	X0[0]=G;
	for(int i=0;i<3;i++) X0[i+1]=wu.f->midPoint[i];
	return C1Rect2Set(X0+C*midVector(X),C,X-midVector(X));
}

//////////////////////////
// Function: DerivativeOf_h()
// 
// Purpose: This function computes the bound on h'(G), where h is the function from the paper, 
// that measures the distance between the stable and unstable manifold, after the unstable manifold
// comes back to the region D, were we investigate the intersection.
//
// The derivation performed in this function is explained in detail in the 
// "dh_notes.pdf" file.
interval DerivativeOf_h(IVector x_,IVector D_, WuEnclosure &wu1, WuEnclosure &wu2, interval G,interval R)
{
	IMatrix C1(4,4);
	C1[0][0] = 1.0; 
	for(int i=0;i<3;i++)
		for(int k=0;k<3;k++) C1[i+1][k+1]=wu1.f->C[i][k];
	
	IMatrix C2=C1;
	swapColumns(C2,1,3);
		
	IVector x(4);
	x[0]=G;
	for(int i=0;i<3;i++) x[i+1]=x_[i];
	
	IVector D(4);
	D[0]=G;
	for(int i=0;i<3;i++) D[i+1]=D_[i];
	
	string vFieldWu1="par:a,b,F;var:G,X,Y,Z;fun:0.0,Y^2+Z^2+a*X-a*F,-X*Y+b*X*Z+Y-G,-b*X*Y-X*Z+Z;";
	string vFieldWu2="par:a,b,F;var:G,X,Y,Z;fun:0.0,-Y^2-Z^2-a*X+a*F,X*Y-b*X*Z-Y+G,b*X*Y+X*Z-Z;";
	
	IMap f1(vFieldWu1);  f1.setParameter("a",0.25);  f1.setParameter("b",4.0);  f1.setParameter("F",4.0);
	IMap f2(vFieldWu2);  f2.setParameter("a",0.25);  f2.setParameter("b",4.0);  f2.setParameter("F",4.0);
	
	LocalVField h1(f1,x,C1);
	LocalVField h2(f2,x,C2);
	interval M=2.0;
	if(satisfiesExtendedRateConditions(h1[D],M,2)==0) abort("error. Problem with establishing the derivative of wu1 with respect to parameter. Aborting program.");
	if(satisfiesExtendedRateConditions(h2[D],M,3)==0) abort("error. Problem with establishing the derivative of wu2 with respect to parameter. Aborting program.");
	
	IVector dWu1_dtheta(4);
	dWu1_dtheta[0]=1.0;  // theta coordinate
	dWu1_dtheta[1]=0.0;  // x coordinate
	dWu1_dtheta[2]=interval(-1.0,1.0)/M;  // y1 coordinate, which is part of wu
	dWu1_dtheta[3]=interval(-1.0,1.0)/M;  // y2 coordinate, which is part of wu
	
	IMatrix Dw2(1,3);
	Dw2[0][0]=interval(-1.0,1.0)/M;
	Dw2[0][1]=interval(-1.0,1.0)*wu2.L;
	Dw2[0][2]=interval(-1.0,1.0)*wu2.L;
	
	IOdeSolver solver(f1,25);
	ITimeMap Phi(solver);
	
	IVector xu(1);
	xu[0]=-R;
	C1Rect2Set Pu=paramSet(wu1,C1,xu,G);
	IMatrix DPhi(4,4);
	
	Phi(50,Pu,DPhi);
	IVector v(3);
	v[0]=1.0;
	v[1]=((gaussInverseMatrix(C2)*DPhi*C1)*dWu1_dtheta)[1];
	v[2]=((gaussInverseMatrix(C2)*DPhi*C1)*dWu1_dtheta)[2];
	
	return ((gaussInverseMatrix(C2)*DPhi*C1)*dWu1_dtheta)[3] - (Dw2*v)[0];
}

////////////////////////
// Function: plotWs()
//
// Purpose: Plotting the enclosure of the two dimensional stable manifold.
// Note: In the proof, all manifolds are of the form (x,w(x)). We are reversing the 
// roles of the coordinates, since we want to plot (w(y),y).
void plotWs(WuEnclosure &wu2)
{
	ofstream file;
	file.open("pic/Ws.txt");
	int n=100;
	for(int i=0;i<n;i++) 
	{
		for(int k=0;k<n;k++) 
		{
			IVector x(2);
			x[0]=part(wu2.domain[0]/1.01,n,i);
			x[1]=part(wu2.domain[1]/1.01,n,k);
			IVector p=wu2.localImage(x);
			plot(file,p[2],p[1]); // here x is on coordinate [2] since coordinates of Ws are interchanged with respect to Wu
		}
	}
}


///////////////////////
// Function: plotWu()
//
// Purpose: Plotting a point in the local coordinates of the 1-dimensional unstable manifold.
// The point p will be chosen to be Phi(T,(x,wu(x))). In other words, the point p will be the point
// from the unstable manifold, after it returns to the domein D.
void plotWu(IVector p,IMatrix C1,IVector x,string fileName)
{
	ofstream file;
	file.open(fileName);
	p=gauss(C1,p-x);
	plot(file,p[0],p[1]);
}

////////////////////////
// Function: plotWs()
//
// Purpose: Plotting the enclosure of the one dimensional unstable manifold, in its local coordinates.
void plotWu(WuEnclosure &wu1,string fileName)
{
	ofstream file;
	file.open(fileName);
	int n=100;
	for(int i=0;i<n;i++) 
	{
		IVector x(1);
		x[0]=part(wu1.domain[0]/1.01,n,i);
		IVector p=wu1.localImage(x);
		plot(file,p[0],p[1]);
	}
}

///////////////////////////
// Function: Lorenz84()
//
// This is the routine in which the proof is caried out.
void Lorenz84()
{
	// This is the interval of parameters we consider:
	interval G=interval(0.0752761095,0.07527611625);
	
	// This matrix is used to obtain local coordinates of 1-dimensional unstable manifold
	IMatrix C1(3,3);
	C1[0][0]=1.0;                  C1[0][1]=-0.00016604653053618; C1[0][2]=0.00040407899883959;
	C1[1][0]=0.00016384655297642;  C1[1][1]=-0.28235213046095;    C1[1][2]=0.71764786953905;
	C1[2][0]=-0.0011562746220118;  C1[2][1]=0.71764798264861;     C1[2][2]=0.28235189601999;
	
	// The matrix C2 will be chosen to obtain local coordinates of 2-dimensional stable manifold Ws.
	// We establish all manifolds as unstable manifolds, hence for Ws we will reverse the sign
	// of the vector field, and we swap the roles of the coordinates.
	IMatrix C2=C1;
	swapColumns(C2,0,2);
	
	
	// vFieldWu1 is the original vector field.
	// vFieldWu2 is the vector field with reversed sign. It is used to establish the 2-dimensional
	// stable manifold (which becomes 2-dimensional unstable for this vector field).
	string vFieldWu1="par:a,b,F,G;var:X,Y,Z;fun:Y^2+Z^2+a*X-a*F,-X*Y+b*X*Z+Y-G,-b*X*Y-X*Z+Z;";
	string vFieldWu2="par:a,b,F,G;var:X,Y,Z;fun:-Y^2-Z^2-a*X+a*F,X*Y-b*X*Z-Y+G,b*X*Y+X*Z-Z;";
	
	IMap f1(vFieldWu1);  f1.setParameter("a",0.25);  f1.setParameter("b",4.0);  f1.setParameter("F",4.0);
	IMap f2(vFieldWu2);  f2.setParameter("a",0.25);  f2.setParameter("b",4.0);  f2.setParameter("F",4.0);
	
	f2.setParameter("G",G);
	
	// The point x will be the center of our local coordinates. It does not need to enclose any fixed points
	// and can be chosen roughly
	IVector x(3);
	x[0]=4.0;
	x=roughFixedPoint(f2,x);
	
	// Here we compute the bound for all the fixed points. This is a rigorous enclosure.
	f1.setParameter("G",G);
	IVector X=fixedPoint(f1,x);
	cout << "The bound for the fixed point for all G in " << G << " is:" << endl;
	cout << X << endl;
	
	// We now verify that all the fixed points are hyperbolic
	computeEigenvalues(f1,X);
	
	// Our domain in which we enclose the manifolds is D = Bu(R) x Bs(R)
	// Below is the radius R
	interval R=pow(10.0,-4.0);
	
	// -------------------------------
	// Here we establish the 2-dimensional stable manifold 
	// (which is here treated as an unstable manifold for reversed sign vector field)
	
	// The manifold is considered in this local vector field
	LocalVField h2(f2,x,C2);
	
	// This is the two dimensional domain of the stable manifold Ws. We choose so that it contains 
	// the two dimensional ball Bs(R).
	IVector domain2(2);
	domain2[0]=R*interval(-1.0,1.0);
	domain2[1]=R*interval(-1.0,1.0);
	
	// This is the Lipschitz constant for Ws
	interval L2=pow(10.0,-3.0);
	
	// The domain in which we establish Ws is
	//   D = domain2 x Bs
	// Below we choose Bs so that it contains a ball (in fact one dimensional interval) or radius R.
	IVector Bs(1);
	Bs[0]=domain2[0];
	
	// Finally we establish Ws and plot it. While the constructor is initiated, all needed assumptions are checked.
	// if some of them are not satisfied, the constructor will terminate the program and throw an error.
	WuEnclosure wu2(h2,domain2,L2,Bs);
	plotWs(wu2);
	// ----------------------------
	
	// We now turn to the one dimensional unstable manifold.
	// We first obtain its local enclosure. We consider it inside of the set D:
	IVector D(3);
	for(int i=0;i<3;i++) D[i]=R*interval(-1.0,1.0);
	
	// We will divide the parameter range E into N intervals and consider each interval separately.
	// For each of the N subintervals we condust the followng steps:
	// 1. compute local enclosure of the unstable manifold inside of D. This is done in local coordinates.
	// 2. Propagate a point 'pu' from the manifold along the flow for time T computing Phi(T,pu). 
	//    Both 'pu' and 'p' are in the original (not local) coordinates
	// 3. transport p to local coordinates, and check that it is in D
	// 4. compute h'(G), for the considered parameter range.
	// 5. Compute h(G) for the lowest and highest considered parameters.
	int N=10;
	interval derivativeBound;
	for(int k=0;k<N+2;k++)
	{
		if(k<N) f1.setParameter("G",part(G,N,k));
		if(k==N) f1.setParameter("G",G.left());
		if(k==N+1) f1.setParameter("G",G.right());
			
		// 1.
		LocalVField h1(f1,x,C1);
		IVector domain1(1);
		domain1[0]=R*interval(-1.0,1.0);
		interval L1=pow(10.0,-5.0);
		WuEnclosure wu1(h1,domain1,L1);
	
		// 2.
		IOdeSolver solver(f1,25);
		ITimeMap Phi(solver);
	
		IVector xu(1);
		xu[0]=-R;
		C0Rect2Set pu=wu1(xu);
		
		IVector p=Phi(50,pu);
		if(k==N) plotWu(p,C1,x,"pic/WuLeft.txt");
		if(k==N) plotWu(wu1,"pic/WuGLeft.txt");
		if(k==N+1) plotWu(p,C1,x,"pic/WuRight.txt");
		if(k==N+1) plotWu(wu1,"pic/WuGRight.txt");
		
		// 3.
		p=gauss(C2,p-x);
		IVector y(2);
		y[0]=p[0];
		y[1]=p[1];
		if(! (p.euclNorm()<R)) abort("Phi(T,pu) is not in D! Aborting program.");
		if(! subsetInterior(p[2],Bs[0])) abort("Phi(T,pu) is not in D! Aborting program.");				
		
		// 4.
		if(k==0) derivativeBound=DerivativeOf_h(x,D,wu1,wu2,part(G,N,k),R);
		if(k<N && k>0) derivativeBound=intervalHull(derivativeBound,DerivativeOf_h(x,D,wu1,wu2,part(G,N,k),R));
		if(k<N) cout << "verified that PhiT(pu,G_" << k << ") is in D " << endl;
		
		// 5.
		if(k==N) cout << "h("<< G.left() << ") = " << (p-wu2.localImage(y))[2] <<  endl;
		if(k==N+1) cout << "h(" << G.right() <<") = " << (p-wu2.localImage(y))[2] <<  endl;	
	}
	cout << "h'(G) is in " << derivativeBound << endl;
}

int main(int argc, char* argv[])
{
    time_t before,after;
    before=time(NULL);
	
  	cout.precision(14);
	try
	{
		Lorenz84();
	}
	catch(exception& e)
  	{
    		cout << "\n\nException caught: "<< e.what() << endl;
  	}
    
	after=time(NULL);
    cout << "time: " << difftime(after,before) << " s. " << endl;
  	return 0;
} 
